'''
本模块储存着几种常用坐标变换

This module stores several common coordinate transformations
'''
import numpy as np


def dirac(x, num=None, rl=0):
    '''
    将狄拉克算符转换为向量
    
    参数
    ----
    x：一维数组类型，数据
    num：一维数组类型，代表每一个量子位的进制，默认均为二进制
    rl：rl=0或rl='r'代表右矢，rl=1或rl='1'代表左矢
    
    返回
    ----
    二维ndarray，狄拉克算符的向量形式，右矢为列向量，左矢为行向量
    
    
    Convert Dirac Operator to Vector
    
    Parameters
    ----------
    x: 1D-array, data
    num: 1D-array, represents the base of each qubit, which is binary by default
    rl: rl=0 or rl='r' means right vector, rl=1 or rl='l' means left vector
    
    Return
    ------
    2D-ndarray, the vector form of Dirac operator, the right vector is the column vector and the left vector is the row vector
    '''
    x = np.array(x)
    total = x.shape[0]
    if num is None:
        num = np.ones_like(x) * 2
    n = 1
    for i in num:
        n *= i
    m = x[-1]
    for i in range(1, total):
        m += x[- i - 1] * np.prod(num[-i:])
    if rl == 0 or rl == 'r':
        vec = np.zeros((n, 1))
        vec[m, 0] = 1
    elif rl == 1 or rl == 'l':
        vec = np.zeros((1, n))
        vec[0, m] = 1
    return vec


def lorentz(v, x):
    '''
    洛伦兹坐标变换
    
    参数
    ----
    v：惯性系的相对速度，当速度方向沿着x轴正方向时（正变换）v为正
    x：一维或二维列表，(x, y, z, t)
    
    返回
    ----
    ndarray，转换后的坐标
    
    
    Lorentz coordinate transformation, S ←→ S'
    
    Parameters
    ----------
    v: relative velocity of inertial system, 'v' is positive when the velocity direction is along the positive direction of x-axis (positive transformation)
    x: 1-D or 2-D list, (x, y, z, t)
    
    Return
    ------
    ndarray, converted coordinates
    '''
    x = np.array(x, dtype=np.float)
    γ = 1 / (1 - (v / 299792458.0)**2)**0.5
    if len(x.shape) == 1:
        xp = γ * (x[0] - v * x[3])
        tp = γ * (x[3] - v * x[0] / 299792458.0**2)
        x[0], x[3] = xp, tp
    
    else:
        xp = γ * (x[:, 0] - v * x[:, 3])
        tp = γ * (x[:, 3] - v * x[:, 0] / 299792458.0**2)
        x[:, 0], x[:, 3] = xp, tp
    
    return x
    

def lorentz_v(v, vo):
    '''
    洛伦兹速度变换
    
    参数
    ----
    v：惯性系的相对速度，当速度方向沿着x轴正方向时（正变换）v为正
    vo：一维或二维列表，(vx, vy, vz)
    
    返回
    ----
    ndarray，转换后的速度
    
    
    Lorentz speed transformation, S ←→ S'
    
    Parameters
    ----------
    v: relative velocity of inertial system, 'v' is positive when the velocity direction is along the positive direction of x-axis (positive transformation)
    vo: 1-D or 2-D list, (vx, vy, vz)
    
    Return
    ------
    ndarray, converted velocity
    '''
    vo = np.array(vo, dtype=np.float)
    γ = 1 / (1 - (v / 299792458.0)**2)**0.5
    if len(vo.shape) == 1:
        factor = 1 - v * vo[0] / 299792458.0**2
        vo[0] = (vo[0] - v) / factor
        vo[1] = vo[1] * γ / factor
        vo[2] = vo[2] * γ / factor
    
    else:
        factor = 1 - v * vo[:, 0] / 299792458.0**2
        vo[:, 0] = (vo[:, 0] - v) / factor
        vo[:, 1] = vo[:, 1] * γ / factor
        vo[:, 2] = vo[:, 2] * γ / factor
    
    return vo


def polar(x):
    '''
    极坐标或柱坐标正变换
    
    参数
    ----
    x：一维或二维列表，(x, y)或(x, y, z)
    
    返回
    ----
    ndarray，转换后的坐标
    
    
    Polar or Cylinder positive transformation
    
    Parameters
    ----------
    x: 1-D or 2-D list, (x, y) or (x, y, z)
    
    Return
    ------
    ndarray, converted coordinates
    '''
    x = np.array(x, dtype=np.float)
    if len(x.shape) == 1:
        r = (x[0]**2 + x[1]**2)**0.5
        if x[0] == 0:
            if x[1] > 0:
                theta = np.pi/2
            elif x[1] < 0:
                theta = -np.pi/2
            elif x[1] == 0:
                theta = 0
        else:
            theta = np.arctan(x[1] / x[0])
            if x[0] < 0 and x[1] > 0:
                theta += np.pi
            elif x[0] < 0 and x[1] < 0:
                theta -= np.pi
        x[0], x[1] = r, theta
    
    else:
        r = (x[:, 0]**2 + x[:, 1]**2)**0.5
        index1 = np.where(x[:, 0] != 0)[0]
        index2 = np.where(x[:, 0] == 0)[0]
        index3 = np.all([x[:, 0] < 0, x[:, 1] > 0], axis=0)
        index3 = np.where(index3 == True)[0]
        index4 = np.all([x[:, 0] < 0, x[:, 1] < 0], axis=0)
        index4 = np.where(index4 == True)[0]
        
        x[index1, 1] = np.arctan(x[index1, 1] / x[index1, 0])
        x[index3, 1] += np.pi
        x[index4, 1] -= np.pi
        
        x_new = x[index2, 1]
        x_new[np.where(x_new > 0)[0]] = np.pi/2
        x_new[np.where(x_new < 0)[0]] = -np.pi/2
        x[index2, 1] = x_new
        x[:, 0] = r
    
    return x


def ipolar(x):
    '''
    极坐标或柱坐标逆变换
    
    参数
    ----
    x：一维或二维列表，(r, θ)或(r, θ, z)
    
    返回
    ----
    ndarray，转换后的坐标
    
    
    Polar or Cylinder inverse transformation
    
    Parameters
    ----------
    x: 1-D or 2-D list, (r, θ) or (r, θ, z)
    
    Return
    ------
    ndarray, converted coordinates
    '''
    x=np.array(x, dtype=np.float)
    if len(x.shape) == 1:
        x0 = x[0] * np.cos(x[1])
        x1 = x[0] * np.sin(x[1])
        x[0], x[1] = x0, x1
    
    else:
        x0 = x[:, 0] * np.cos(x[:, 1])
        x1 = x[:, 0] * np.sin(x[:, 1])
        x[:, 0], x[:, 1] = x0, x1
    
    return x


def sphere(x):
    '''
    球坐标正变换
    
    参数
    ----
    x：一维或二维列表，(x, y, z)
    
    返回
    ----
    ndarray，转换后的坐标
    
    
    Sphere positive transformation
    
    Parameters
    ----------
    x: 1-D or 2-D list, (x, y, z)
    
    Return
    ------
    ndarray, converted coordinates
    '''
    x = np.array(x, dtype=np.float)
    if len(x.shape) == 1:
        r = (x[0]**2 + x[1]**2 + x[2]**2)**0.5
        if r == 0:
            theta = 0
            phai = 0
        else:
            phai = np.arccos(x[2] / r)
            if x[0] == 0:
                if x[1] > 0:
                    theta = np.pi/2
                elif x[1] < 0:
                    theta = -np.pi/2
                elif x[1] == 0:
                    theta = 0
            else:
                theta = np.arctan(x[1] / x[0])
        x[0], x[1], x[2] = r, theta, phai
    
    else:
        r = (x[:, 0]**2 + x[:, 1]**2 + x[:, 2]**2)**0.5
        index_r1 = np.where(r == 0)[0]
        x[index_r1] = 0
        
        index_r2 = np.where(r != 0)[0]
        x_new = x[index_r2]
        r_new = r[index_r2]
        x_new[:, 2] = np.arccos(x_new[:, 2] / r_new)    # 处理了phi
        
        index1 = np.where(x_new[:, 0] != 0)[0]
        index2 = np.where(x_new[:, 0] == 0)[0]
        
        x_new[index1, 1] = np.arctan(x_new[index1, 1] / x_new[index1, 0])
        x_new_new = x_new[index2, 1]
        x_new_new[np.where(x_new_new > 0)] = np.pi/2
        x_new_new[np.where(x_new_new < 0)] = -np.pi/2
        x_new[index2, 1] = x_new_new
        x_new[:, 0] = r_new
        x[index_r2] = x_new             # 处理了theta和r
    
    return x


def isphere(x):
    '''
    球坐标逆变换
    
    参数
    ----
    x：一维或二维列表，(r, θ, φ)
    
    返回
    ----
    ndarray，转换后的坐标
    
    
    Sphere inverse transformation
    
    Parameters
    ----------
    x: 1-D or 2-D list, (r, θ, φ)
    
    Return
    ------
    ndarray, converted coordinates
    '''
    x = np.array(x)
    if len(x.shape) == 1:
        x0 = x[0] * np.sin(x[2]) * np.cos(x[1])
        x1 = x[0] * np.sin(x[2]) * np.sin(x[1])
        x2 = x[0] * np.cos(x[2])
        x[0], x[1], x[2] = x0, x1, x2
    
    else:
        x0 = x[:, 0] * np.sin(x[:, 2]) * np.cos(x[:, 1])
        x1 = x[:, 0] * np.sin(x[:, 2]) * np.sin(x[:, 1])
        x2 = x[:, 0] * np.cos(x[:, 2])
        x[:, 0], x[:, 1], x[:, 2] = x0, x1, x2
    
    return x


def rotate(theta, x):
    '''
    平面直角坐标系的旋转变换
    逆时针旋转时theta为正，顺时针旋转时theta为负
    
    参数
    ----
    x：一维或二维列表，(x, y)
    theta：浮点数类型，坐标系绕原点逆时针旋转的角度
    
    返回
    ----
    ndarray，转换后的坐标
    
    
    Rotation transformation of plane rectangular coordinate system
    'theta' is positive when rotating anticlockwise and negative when rotating clockwise
    
    Parameter
    ---------
    x: 1-D or 2-D list, (x, y)
    theta: float, the angle that the coordinate system rotates counterclockwise about the origin
    
    Return
    ------
    ndarray, converted coordinates
    '''
    x = np.array(x, dtype=np.float)
    if len(x.shape) == 1:
        x0 = np.cos(theta) * x[0] + np.sin(theta) * x[1]
        x1 = np.cos(theta) * x[1] - np.sin(theta) * x[0]
        x[0], x[1] = x0, x1
    
    else:
        x0 = np.cos(theta) * x[:, 0] + np.sin(theta) * x[:, 1]
        x1 = np.cos(theta) * x[:, 1] - np.sin(theta) * x[:, 0]
        x[:, 0], x[:, 1] = x0, x1
    
    return x


'''
该模块用于统计分析

This module is uesd for statistical analysis
'''
import numpy as np
from matplotlib import pyplot as plt
from maysics.calc import inte
from maysics.utils import grid_net
from scipy.stats import chi2
from scipy.interpolate import interp1d


def r_moment(data, p_range=None, args={}, k=1, acc=0.1):
    '''
    计算原点矩
    
    参数
    ----
    data：一维列表或函数类型，样本或概率密度函数
    p_range：元组类型，可选，计算上下限
    args：字典类型，可选，用于传递func中的其他参数
    k：整型，可选，阶数，默认为1
    acc：浮点数类型，可选，积分精度，仅在data为函数类型时有效，默认为0.1
    
    返回
    ----
    数，原点矩
    
    
    Calculate raw moment
    
    Parameters
    ----------
    fdata: 1D list or function, a sample or probability density function
    p_range：tuple, callable, lower and upper limit of calculation
    args: dict, callable, pass other parameters to func
    k: int, callable, orders, default=1
    acc: float, callable, integration accuracy, it's effective only when data is function, default=0.1
    
    Return
    ------
    num, raw moment
    '''
    if type(data).__name__ == 'function' or type(data).__name__ == 'method':
        def ex_func(x):
            return x**k * data(x, **args)
        
        return inte(ex_func, [p_range], acc=acc)
    
    else:
        data = np.array(data)**k
        return data.mean()


def ex(data, p_range=None, args={}, acc=0.1):
    '''
    计算数学期望，等价于一阶原点矩
    
    参数
    ----
    data：一维列表或函数类型，样本或概率密度函数
    p_range：元组类型，可选，计算上下限
    args：字典类型，可选，用于传递func中的其他参数
    acc：浮点数类型，可选，积分精度，仅在data为函数类型时有效，默认为0.1
    
    返回
    ----
    数，数学期望
    
    
    Calculate mathematical expectation, which is equivalent to one order raw moment
    
    Parameters
    ----------
    fdata: 1D list or function, a sample or probability density function
    p_range：tuple, callable, lower and upper limit of calculation
    args: dict, callable, pass other parameters to func
    acc: float, callable, integration accuracy, it's effective only when data is function, default=0.1
    
    Return
    ------
    num, mathematical expectation
    '''
    if type(data).__name__ == 'function' or type(data).__name__ == 'method':
        def ex_func(x):
            return x * data(x, **args)
        
        return inte(ex_func, [p_range], acc=acc)
    
    else:
        data = np.array(data)
        return data.mean()


def c_moment(data, p_range=None, args={}, k=1, acc=0.1):
    '''
    计算中心矩
    
    参数
    ----
    data：一维列表或函数类型，样本或概率密度函数
    p_range：元组类型，可选，计算上下限
    args：字典类型，可选，用于传递func中的其他参数
    k：整型，可选，阶数，默认为1
    acc：浮点数类型，可选，积分精度，仅在data为函数类型时有效，默认为0.1
    
    返回
    ----
    数，中心矩
    
    
    Calculate central moment
    
    Parameters
    ----------
    data: 1D list or function, a sample or probability density function
    p_range：tuple, callable, lower and upper limit of calculation
    args: dict, callable, pass other parameters to func
    k: int, callable, orders, default=1
    acc: float, callable, integration accuracy, it's effective only when data is function, default=0.1
    
    Return
    ------
    num, central moment
    '''
    raw_exp = ex(data, p_range, args)
    
    if type(data).__name__ == 'function' or type(data).__name__ == 'method':
        def ex_func(x):
            return (x-raw_exp)**k * data(x, **args)
        
        return inte(ex_func, [p_range], acc=acc)
    
    else:
        data = np.array(data) - raw_exp
        return r_moment(data, k=k)


def dx(data, p_range=None, args={}, acc=0.1):
    '''
    计算方差，等价于二阶中心矩
    
    参数
    ----
    data：一维列表或函数类型，样本或概率密度函数
    p_range：元组类型，可选，计算上下限
    args：字典类型，可选，用于传递func中的其他参数
    acc：浮点数类型，可选，积分精度，仅在data为函数类型时有效，默认为0.1
    
    返回
    ----
    数，方差
    
    
    Calculate variance, which is equivalent to two orders central moment
    
    Parameters
    ----------
    data: 1D list or function, a sample or probability density function
    p_range：tuple, callable, lower and upper limit of calculation
    args: dict, callable, pass other parameters to func
    acc: float, callable, integration accuracy, it's effective only when data is function, default=0.1
    
    Return
    ------
    num, variance
    '''
    return c_moment(data, p_range, args, 2, acc)


def skew(data, p_range=None, args={}, acc=0.1):
    '''
    计算偏度，等价于三阶中心矩
    
    参数
    ----
    data：一维列表或函数类型，样本或概率密度函数
    p_range：元组类型，可选，计算上下限
    args：字典类型，可选，用于传递func中的其他参数
    acc：浮点数类型，可选，积分精度，仅在data为函数类型时有效，默认为0.1
    
    返回
    ----
    数，偏度
    
    
    Calculate skewness, which is equivalent to three orders central moment
    
    Parameters
    ----------
    data: 1D list or function, a sample or probability density function
    p_range：tuple, callable, lower and upper limit of calculation
    args: dict, callable, pass other parameters to func
    acc: float, callable, integration accuracy, it's effective only when data is function, default=0.1
    
    Return
    ------
    num, skewness
    '''
    return c_moment(data, p_range, args, 3, acc)


def kurt(data, p_range=None, args={}, acc=0.1):
    '''
    计算峰度，等价于四阶中心矩
    
    参数
    ----
    data：一维列表或函数类型，样本或概率密度函数
    p_range：元组类型，可选，计算上下限
    args：字典类型，可选，用于传递func中的其他参数
    acc：浮点数类型，可选，积分精度，仅在data为函数类型时有效，默认为0.1
    
    返回
    ----
    数，峰度
    
    
    Calculate kurtosis, which is equivalent to four orders central moment
    
    Parameters
    ----------
    data: 1D list or function, a sample or probability density function
    p_range：tuple, callable, lower and upper limit of calculation
    args: dict, callable, pass other parameters to func
    acc: float, callable, integration accuracy, it's effective only when data is function, default=0.1
    
    Return
    ------
    num, kurtosis
    '''
    return c_moment(data, p_range, args, 4, acc)


def mle(func, data, p_range, acc=0.1):
    '''
    最大似然法
    求解似然函数L(θ) = ln(∏ f(xi; θ)) = ∑ ln(f(xi; θ))在θ ∈ p_range范围内的最大值

    参数
    ----
    func：函数类型，密度概率函数，func的形式应如下：
        def func(x, param):
        其中x需输入一个数或一个向量，是随机变量；param是一维ndarray，为未确定的参数
    data：一维或二维数组，样本
    p_range：二维数组，未确定参数的取值范围
    acc：数类型，可选，精度，默认为0.1
    
    返回
    ----
    1D ndarray，最佳的参数值


    Maximum Likelihood Estimate
    Find the maximum of likelihood function L(θ) = ln(∏ f(xi; θ)) = ∑ ln(f(xi; θ)) when θ ∈ p_range

    Parameters
    ----------
    func：function, probability density function, the form of func should be as follows：
        def func(x, param):
        x is a number or a vector, random variable; param is 1D ndarray, undetermined parameters
    data：1D or 2D array，samples
    p_range：2D array, callable, value range of undetermined parameters
    acc：num, callable, accuracy, default=0.1
    
    Return
    ------
    1D ndarray, the best parameter value
    '''
    # 构建似然函数的相反数
    def Lmin(theta, data):
        result = 0
        for i in data:
            single_result = func(i, theta)
            if single_result <= 0:
                return float('-inf')
            result += np.log(single_result)
        return result
    
    p_range = np.array(p_range)
    p_range_net = []
    for i in p_range:
        p_range_net.append(np.arange(i[0], i[1], acc))
    
    p_range_net = grid_net(*p_range_net)
    
    y = []
    for i in p_range_net:
        y.append(Lmin(i, data))
    y = np.array(y)
    x = np.argmax(y)
    
    return p_range_net[x]


def inde_test(data):
    '''
    独立性检验
    
    参数
    ----
    data：二维数组
    
    返回
    ----
    接受假设的错误概率，即认为两变量没有关系的错误概率
    
    
    Independence Test
    
    Parameter
    ---------
    data: 2D array
    
    Return
    ------
    the error probability of accepting the hypothesis (the error probability that the two variables are not related)
    '''
    data = np.array(data, dtype=float)
    data_shape = data.shape
    sum_1 = np.array([data.sum(axis=1)]).T
    sum_2 = data.sum(axis=0)
    p1 = data / sum_1
    p2 = data / sum_2
    sum_3 = data.sum()
    p_data = p1 * p2 * sum_3
    data = (data - p_data)**2 / p_data
    data = data.sum()
    
    return chi2.cdf(data, (data_shape[0] - 1) * (data_shape[1] - 1))


class DF1d():
    '''
    一维分布拟合
    用频率分布函数逼近概率密度函数
    
    参数
    ----
    sample：1D列表，样本点
    span：1D数组，区间间隔，如span = [a, b, c]则将区间分为[a, b]和[b, c]，并统计各区间频率
    kind：浮点数类型或整型，可选，将插值类型指定为字符串('linear'、'nearest'、'zero'、'slinear'、'squardic'、'previous'、'next'，其中'zero'、'slinear'、'squared'和'cubic'表示零阶、一阶、二阶或三阶样条曲线插值；'previous'和'next'只返回点的上一个或下一个值)或作为一个整数指定要使用的样条曲线插值器的顺序。默认为'linear'
    
    属性
    ----
    f：概率密度函数
    
    
    1D distribution fitting
    Approximation of probability density function by frequency distribution function
    
    Parameters
    ----------
    sample: 1D list, samples
    span: 1D array, span of intervals, e.g. span = [a, b, c] will divide the interval into [a, b] and [b, c], and count the frequency of each interval
    kind: float or int, callable, specifies the kind of interpolation as a string ('linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic', 'previous', 'next', where 'zero', 'slinear', 'quadratic' and 'cubic' refer to a spline interpolation of zeroth, first, second or third order; 'previous' and 'next' simply return the previous or next value of the point) or as an integer specifying the order of the spline interpolator to use. default='linear'
    
    Attribution
    -----------
    f: probability density function
    '''
    def __init__(self, sample, span, kind='linear'):
        x_list = []
        y_list = []
        sample = np.array(sample)
        total_num = len(sample)
        for i in range(len(span)-1):
            num = len(sample[sample < span[i+1]]) - len(sample[sample < span[i]])
            y_list.append(num / total_num)
            x_list.append(0.5 * (span[i] + span[i+1]))
        self.f = interp1d(x_list, y_list, kind=kind)
        self.__minmax = [min(x_list), max(x_list)]
    
    
    def show(self, acc=0.01):
        '''
        作图并显示
        
        参数
        ----
        acc：浮点数类型，精度
        
        
        Display the image
        
        Parameters
        ----------
        acc: float, accuracy
        '''
        x = np.arange(self.__minmax[0], self.__minmax[1], acc)
        plt.plot(x, self.f(x))
        plt.show()
    
    
    def savefig(self, filename, acc=0.01):
        '''
        作图并保存
        
        参数
        ----
        filename：字符串类型，保存的文件名
        acc：浮点数类型，精度
        
        
        Save the image
        
        Parameters
        ----------
        filename: str, file name
        acc: float, accuracy
        '''
        x = np.arange(self.__minmax[0], self.__minmax[1], acc)
        plt.plot(x, self.f(x))
        plt.savefig(filename)


class DFT():
    '''
    单个分布拟合检验
    
    参数
    ----
    func_type：字符串类型，分布拟合采用的函数的类型，可选'pdf'(概率密度函数)、'cdf'(分布函数)和'pmf'(离散分布函数)，默认为'pdf'
    
    属性
    ----
    degree：卡方分布的自由度
    chi2_value：卡方值
    P：拒绝假设的错误概率
    
    
    Single distribution fitting test

    Parameter
    ---------
    func_type: str, callable, types of functions used in distribution fitting, 'pdf'(probability density function), 'cdf'(cumulative distribution function) and 'pmf'(discrete distribution function) are optional, default='pdf'
    
    Atrributes
    ----------
    degree: degrees of freedom of chi square distribution
    chi2_value: chi square value
    P: Probability of error in rejecting hypothesis
    '''
    def __init__(self, func_type='pdf'):
        self.__func_type = func_type
    
    
    def __merge(self, data_list, pro_list):
        # 找到npi < 5的元素位置
        delete_index = np.where(pro_list < 5)[0]
        if len(delete_index) > 0:
            delete_pro = pro_list[delete_index]
            delete_data = data_list[delete_index]
            pro_final = [delete_pro[0]]
            data_final = [delete_data[0]]
            for i in range(1, len(delete_pro)):
                if pro_final[-1] < 5:
                    pro_final[-1] += delete_pro[i]
                    data_final[-1] += delete_data[i]
                else:
                    pro_final.append(delete_pro[i])
                    data_final.append(delete_data[i])
            pro_final = np.array(pro_final)
            data_final = np.array(data_final)
            
            # 删除原列表npi < 5位置对应的元素
            data_list = np.delete(data_list, delete_index)
            pro_list = np.delete(pro_list, delete_index)
            
            data_list = np.hstack((data_list, data_final))
            pro_list = np.hstack((pro_list, pro_final))
            
            if len(pro_list) == 1:
                raise Exception("Degree of freedom of chi-square distribution is 0.")
            
            elif len(pro_list) > 1 and pro_list[-1] < 5:
                data_list[-2] += data_list[-1]
                pro_list[-2] += pro_list[-1]
                data_list = np.delete(data_list, -1)
                pro_list = np.delete(pro_list, -1)
        
        return data_list, pro_list
    
    
    def __con_fit(self, data, func, num_data, args, acc):
        data = sorted(data)
        
        # 从小到大按20个一组分成多个部分
        data_list = []
        for i in range(int(num_data / 20 + 1)):
            data_list.append(data[i*20: (i+1)*20])
        
        # 如果最后一个部分的数量只有0个或1个元素，则叠加到前面
        if len(data_list[-1]) == 1:
            data_list[-2].append(data_list[-1][0])
            del data_list[-1]
        
        elif len(data_list[-1]) == 0:
            del data_list[-1]
        
        # 计算npi数组
        pro_list = []
        if self.__func_type == 'pdf':
            for i in data_list:
                pro_list.append(inte(func, [[i[0], i[-1]]], args=args, acc=acc))
        
        elif self.__func_type == 'cdf':
            for i in data_list:
                pro_list.append(func(i[-1], **args) - func(i[0], **args))
        
        pro_list = np.array(pro_list) * num_data
        
        # 将数据数组转化为fi数组
        for i in range(len(data_list)):
            data_list[i] = len(data_list[i])
        data_list = np.array(data_list)
        data_list, pro_list = DFT.__merge(self, data_list, pro_list)
        return data_list, pro_list
    

    def __dis_fit(self, data, func, num_data, args):
        data = list(data)
        data_set = sorted(set(data))

        # 求出各data中个元素出现的次数并计算npi数组
        data_list = []
        pro_list = []
        for i in data_set:
            data_list.append(data.count(i))
            pro_list.append(func(i, **args))
        data_list = np.array(data_list)
        pro_list = np.array(pro_list) * num_data

        data_list, pro_list = DFT.__merge(self, data_list, pro_list)

        return data_list, pro_list
    
    
    def fit(self, data, func, args={}, acc=0.1):
        '''
        参数
        ----
        data：数据
        func：函数类型，分布函数或概率密度函数，函数的输入须为一个数
        args：字典类型，可选，用于传递func中的其他参数
        acc：浮点数类型，可选，积分精度，仅func_type为'pdf'或'cdf'时有效，默认为0.1
        
        
        Parameters
        ----------
        data: data
        func: function, cumulative distribution function or probability density function, the input of func should be a number
        args: dict, callable, pass other parameters to func
        acc: float, callable, integration accuracy, it's effective only when func_type is 'pdf' or 'cdf', default=0.1
        '''
        self.__data = np.array(data, dtype=np.float)
        self.__func = func
        self.__args = args
        num_data = len(data)
        
        if self.__func_type != 'pmf':
            data_list, pro_list = self.__con_fit(data, func, num_data, args, acc)
        else:
            data_list, pro_list = self.__dis_fit(data, func, num_data, args)
        
        self.degree = len(pro_list) - 1
        self.chi2_value = sum((data_list - pro_list)**2 / pro_list)
        self.P = chi2.cdf(self.chi2_value, self.degree)
    
    
    def __image_process(self, acc):
        x_min = min(self.__data)
        x_max = max(self.__data)
        if self.__func_type == 'pmf':
            acc = 1
        x = np.arange(x_min, x_max, acc)
        y = self.__func(x, **self.__args)
        
        func_type = type(y).__name__
        if func_type == 'int' or func_type == 'float':
            y = np.ones_like(x) * 0.3
        else:
            y *= (0.3 / y.max())
        
        quartile1, medians, quartile3 = np.percentile(self.__data, [25, 50, 75])
        
        ax = plt.violinplot(self.__data, widths=0.6, showextrema=False, vert=False)
        ax['bodies'][0].set_facecolor('red')
        ax['bodies'][0].set_edgecolor('black')
        ax['bodies'][0].set_alpha(1)
        
        plt.scatter(medians, 1, marker='o', color='white', s=30, zorder=3)
        plt.hlines(1, quartile1, quartile3, color='k', linestyle='-', lw=5)
        plt.hlines(1, x_min, x_max, color='k', linestyle='-', lw=1)
        plt.vlines(self.__data.mean(), 0.9, 1.1, color='yellow', linestyle='-', lw=2)
        
        if self.__func_type != 'pmf':
            plt.plot(x, y)
        else:
            plt.scatter(x, y)
            zeros = np.zeros_like(y)
            plt.vlines(x, zeros, y)
        plt.yticks([])
    
    
    def show(self, acc=0.01):
        '''
        显示数据的小提琴图和假设的分布函数或概率密度图像
        
        参数
        ----
        acc：浮点数类型，可选，画图精度，仅func_type为'pdf'或'cdf'时有效，默认为0.01
        
        
        Display the violin chart of data and the image of hypothetical cumulative distribution function or probability density function
        
        Parameter
        ---------
        acc: float, callable, the accuracy of drawing, it's effective only when method is 'pdf' or 'cdf', default=0.01
        '''
        self.__image_process(acc)
        plt.show()
    
    
    def savefig(self,filename, acc=0.01):
        '''
        储存数据的小提琴图和假设的分布函数或概率密度图像
        
        参数
        ----
        filename：字符串类型，文件名
        acc：浮点数类型，可选，画图精度，仅在func_type为'pdf'或'cdf'时有效，默认为0.01
        
        
        Save the violin chart of data and the image of hypothetical cumulative distribution function or probability density function
        
        Parameters
        ----------
        filename: str, file name
        acc: float, callable, the accuracy of drawing, it's effective only when method is 'pdf' or 'cdf', default=0.01
        '''
        self.__image_process(acc)
        plt.savefig(filename)


'''
本模块用于数字图像处理

This module is used for digital image process
'''
import numpy as np
from matplotlib import pyplot as plt


def l_convert(data, dtype=float):
    '''
    将RGB图像转换为灰度图像
    
    参数
    ----
    data：二维或三维ndarray，图像的张量数据
    dtype：可选，输出图像数据的数据格式，默认为float
    
    返回
    ----
    二维ndarray，灰度图像
    
    
    Transform RGB image into grey image
    
    Parameters
    ----------
    data: 2D or 3D ndarray, tensor of image
    dtype: callable, data format of output image, default=float
    
    Return
    ------
    2D ndarray, grey image
    '''
    data = np.array(data, dtype=float)
    data = data[:, :, 0] * 0.299 + data[:, :, 1] * 0.587 + data[:, :, 2] * 0.114
    data[data>255]=255
    data = np.around(data)
    return np.array(data, dtype=dtype)


def hsv(pic, dtype=float):
    '''
    将RGB格式转换为HSV格式
    
    参数
    ----
    data：三维ndarray，图像的RGB张量数据
    dtype：可选，输出图像数据的数据格式，默认为float
    
    返回
    ----
    三维ndarray，HSV数据
    
    
    Transform RGB into HSV
    
    Parameters
    ----------
    data: 3D ndarray, RGB tensor of image
    dtype: callable, data format of output image, default=float
    
    Return
    ------
    3D ndarray, HSV data
    '''
    pic = np.array(pic, dtype=float)
    v = pic.max(axis=2)
    med_v = pic.min(axis=2)
    s = np.zeros_like(v)
    s[v!=0] = 1 - med_v[v!=0] / v[v!=0]
    h = np.zeros_like(v)
    med_v = v - med_v
    
    for i in range(3):
        loc = np.where((v==pic[:, :, i]) & (med_v!=0))
        h[loc] = (pic[:, :, (i+1)%3][loc] - pic[:, :, (i+2)%3][loc]) * 60 / med_v[loc] + i * 120
    h[h<0] += 360
    
    pic = np.array([h, s, v], dtype=dtype)
    return pic.transpose(1, 2, 0)


def ihsv(pic, dtype=float):
    '''
    将HSV格式转换为RGB格式
    
    参数
    ----
    data：三维ndarray，图像的HSV张量数据
    dtype：可选，输出图像数据的数据格式，默认为float
    
    返回
    ----
    三维ndarray，HSV数据
    
    
    Transform HSV into RGB
    
    Parameters
    ----------
    data: 3D ndarray, HSV tensor of image
    dtype: callable, data format of output image, default=float
    
    Return
    ------
    3D ndarray, RGB data
    '''
    pic = np.array(pic, dtype=float)
    h = pic[:, :, 0] / 60
    f = h.copy()
    h = h.astype(int) % 6
    f -= h
    v = [pic[:, :, 2] * (1 - pic[:, :, 1]),
         pic[:, :, 2] * (1 - (1 - f) * pic[:, :, 1]),
         pic[:, :, 2],
         pic[:, :, 2] * (1 - f * pic[:, :, 1])]
    
    # 初始化
    r = np.zeros_like(h)
    g = np.zeros_like(h)
    b = np.zeros_like(h)
    
    # 分别处理h为偶数和奇数时的rgb
    for i in range(3):
        loc = np.where(h==2 * i)
        r[loc] = v[(i+2)%3][loc]
        g[loc] = v[(i+1)%3][loc]
        b[loc] = v[i%3][loc]
    del v[1]
    for i in range(3):
        loc = np.where(h==2 * i + 1)
        r[loc] = v[(i+2)%3][loc]
        g[loc] = v[(i+1)%3][loc]
        b[loc] = v[i%3][loc]
    
    pic = np.array([r, g, b], dtype=dtype)
    return pic.transpose(1, 2, 0)


def hist(data):
    '''
    图像的直方图
    
    参数
    ----
    data：二维或三维ndarray，图像的张量数据
    
    返回
    ----
    一维或二维ndarray，图像的直方图数据
    
    
    Histogram of Image
    
    Parameter
    ---------
    data: 2D or 3D ndarray, tensor of image
    
    Return
    ------
    1D or 2D ndarray, data of hitogram
    '''
    data = np.array(data)
    if len(data.shape) == 2:
        result = []
        for i in range(256):
            result.append(data[data==i].shape[0])
    
    elif len(data.shape) == 3:
        result = [[],[],[]]
        for j in range(3):
            data_copy = data[:, :, j]
            for i in range(256):
                result[j].append(data_copy[data_copy==i].shape[0])
    
    return np.array(result)


def hist_graph(data, mode=1, save=False):
    '''
    绘制图像的直方图
    
    参数
    ----
    data：二维或三维ndarray，图像的张量数据
    mode：数类型，可选1和2，1代表折线图，2代表直方图，默认为1
    save：字符串类型或布尔类型，若为字符串类型，表示保存图像为文件，若为False，则表示显示图像，默认为False
    
    
    Display the Histogram of Image
    
    Parameters
    ----------
    data: 2D or 3D ndarray, tensor of image
    mode: num, 1 and 2 are optional, 1 means line graph, 2 means histogram
    save: str or bool, str type means save graph as file, False means display graph, default=False
    '''
    data = hist(data)
    x = np.arange(0, 256, 1)
    if mode == 1 or mode == 'plot':
        if len(data.shape) == 1:
            plt.plot(x, data)
        
        elif len(data.shape) == 2:
            plt.plot(x, data[0], color='r')
            plt.plot(x, data[1], color='g')
            plt.plot(x, data[2], color='b')
    
    elif mode == 2 or mode == 'bar':
        if len(data.shape) == 1:
            plt.bar(x, data, width=0.5)
        
        elif len(data.shape) == 2:
            fig = plt.figure()
            colors = ['r', 'g', 'b']
            for i in range(3):
                ax = fig.add_subplot(3, 1, i+1)
                ax.bar(x, data[i], width=0.5, color=colors[i])
    
    plt.tight_layout()
    if save is False:
        plt.show()
    else:
        plt.savefig(save)


def _pre_linear_enhancement(data, begin, end, k, center):
    if not center:
        center = (begin + end) * 0.5
    inter1 = k * (begin - center) + center
    inter2 = k * (end - center) + center
    
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            if begin <= data[i, j] <= end:
                data[i, j] = k * (data[i, j] - center) + center
            elif data[i, j] < begin:
                data[i, j] = inter1 * data[i, j] / begin
            elif data[i, j] > end:
                data[i, j] = (255 - inter2) * (data[i, j] - end) / (255 - end) + inter2
    
    return data


def linear_enhance(data, begin, end, k, center=None, dtype=float):
    '''
    线性增强
    
    参数
    ----
    data：二维或三维ndarray，图像的张量数据
    begin：整型，线性增强区域的起点
    end：整型，线性增强区域的终点
    k：数类型，线性增强区域的斜率
    center：线性增强区域的不变点
    dtype：可选，输出图像数据的数据格式，默认为float
    
    返回
    ----
    二维或三维ndarray，图像的张量数据
    
    
    Linear Enhancement
    
    Parameters
    ----------
    data: 2D or 3D ndarray, tensor of image
    begin: starting point of linear enhancement region
    end: end point of linear enhancement region
    k: slope of linear enhancement region
    center: invariant point of linear enhancement region
    dtype: callable, data format of output image, default=float
    
    Return
    ------
    2D or 3D ndarray, tensor of image
    '''
    data = np.array(data, dtype=float)
    if len(data.shape) == 2:
        data_new = _pre_linear_enhancement(data, begin, end, k, center)
    elif len(data.shape) == 3:
        data_new = []
        data = data.transpose(2, 0, 1)
        for i in data:
            data_new.append(_pre_linear_enhancement(i, begin, end, k, center))
        data_new = np.array(data_new).transpose(1, 2, 0)
    
    data_new = np.around(data_new)
    data_new[data_new > 255] = 255
    data_new[data_new < 0] = 0
    return np.array(data_new, dtype=dtype)


def hist_equa(data, dtype=float):
    '''
    直方图均衡
    
    参数
    ----
    data：二维或三维ndarray，图像的张量数据
    dtype：可选，输出图像数据的数据格式，默认为float
    
    返回
    ----
    二维或三维ndarray，图像的张量数据
    
    
    Parameters
    ----------
    data: 2D or 3D ndarray, tensor of image
    dtype: callable, data format of output image, default=float
    
    Return
    ------
    2D or 3D ndarray, tensor of image
    '''
    data = np.array(data, dtype=float)
    data_copy = data.copy()
    if len(data.shape) == 3:
        data_copy = data_copy[:, :, 0] * 0.299 + data_copy[:, :, 1] * 0.587 + data_copy[:, :, 2] * 0.114
        data_copy = np.around(data_copy)
    
    N = np.prod(data_copy.shape)
    grey = []
    for i in range(256):
        grey.append(data_copy[data_copy==i].shape[0])
        data[data==i] = sum(grey) * 255 / N
    
    data = np.around(data)
    return np.array(data, dtype=dtype)


def laplace(data, mode=8, strenth=1, dtype=float):
    '''
    图像的拉普拉斯算子
    
    参数
    ----
    data：二维或三维ndarray，图像的张量数据
    mode：数类型，可选4和8，分别表示两种掩膜：
        [[0, -1, 0],
         [-1, 4, -1],
         [0, -1, 0]]，
         和
        [[-1, -1, -1],
         [-1, 8, -1],
         [-1, -1, -1]]
        默认为8
    strenth：数类型，可选，对比的增强倍率，默认为1
    dtype：可选，输出图像数据的数据格式，默认为float
    
    返回
    ----
    二维ndarray，灰度图像
    
    
    Laplace Operator of Image
    
    Parameters
    ----------
    data: 2D or 3D ndarray, tensor of image
    mode: num, 4 and 8 are optional, represented two masks respectively:
        [[0, -1, 0],
         [-1, 4, -1],
         [0, -1, 0]]，
         and
        [[-1, -1, -1],
         [-1, 8, -1],
         [-1, -1, -1]]
        defualt=8
    strenth: num, callable, contrast enhancement magnification, default=1
    dtype: callable, data format of output image, default=float
    
    Return
    ------
    2D ndarray, grey image
    '''
    data = np.array(data, dtype=float)
    if len(data.shape) == 3:
        data = data[:, :, 0] * 0.299 + data[:, :, 1] * 0.587 + data[:, :, 2] * 0.114
    
    data_copy = data.copy()
    data_new = np.zeros_like(data)
    
    data_copy = np.hstack((data_copy[:, 0:1], data_copy[:, 0:-1]))
    data_new += data - data_copy
    data_copy = data
    
    data_copy = np.hstack((data_copy[:, 0:-1], data_copy[:, -1:]))
    data_new += data - data_copy
    data_copy = data
    
    data_copy = np.vstack((data_copy[0:1, :], data_copy[0:-1, :]))
    data_new += data - data_copy
    data_copy = data
    
    data_copy = np.vstack((data_copy[0:-1, :], data_copy[-1:, :]))
    data_new += data - data_copy
    data_copy = data
    
    if mode == 8:
        data_copy = np.hstack((data_copy[1:, 0:1], data_copy[:-1, :]))
        data_copy = np.vstack((data[0:1, :], data_copy[:, :-1]))
        data_new += data - data_copy
        data_copy = data
        
        data_copy = np.hstack((data_copy[:-1, 0:1], data_copy[1:, :]))
        data_copy = np.vstack((data_copy[:, :-1], data[-2:-1, :]))
        data_new += data - data_copy
        data_copy = data
        
        data_copy = np.hstack((data_copy[:-1, :], data_copy[1:, -2:-1]))
        data_copy = np.vstack((data[0:1, :], data_copy[:, 1:]))
        data_new += data - data_copy
        data_copy = data
        
        data_copy = np.hstack((data_copy[1:, :], data_copy[:-1, -2:-1]))
        data_copy = np.vstack((data_copy[:, 1:], data[-2:-1, :]))
        data_new += data - data_copy
        data_copy = data
    
    data_new = abs(data_new) * strenth
    data_new[data_new > 255] = 255
    return np.around(data_new).astype(dtype)


def saturate(data, scale=1, param={}, dtype=float):
    '''
    调整图片饱和度
    
    参数
    ----
    data：三维ndarray，图像的张量数据
    scale：数或函数类型，可选，当scale为数类型时，表示饱和度调整为S*scale；当scale为函数类型时，表示饱和度调整为scale(S)，默认为1
    param：字典类型，可选，当scale为函数类型且有其他非默认参数时，需输入以参数名为键，参数值为值的字典，默认为空字典
    dtype：可选，输出图像数据的数据格式，默认为float
    
    返回
    ----
    三维ndarray，图像
    
    
    Adjust the saturation
    
    Parameters
    ----------
    data: 3D ndarray, tensor of image
    scale: num or function, callable, when scale is num, it means adjust the saturation to S*scale; while it's function, it means adjust the saturation to scale(S), default=1
    param: dict, callable, When step is function and has other non-default parameters, "param" needs to be input a dictionary with parm_name as key and param_value as value, default={}
    dtype: callable, data format of output image, default=float
    
    Return
    ------
    3D ndarray, image
    '''
    data = np.array(data, dtype=float)
    v = data.max(axis=2)
    med_v = data.min(axis=2)
    s = np.zeros_like(v)
    s[v!=0] = 1 - med_v[v!=0] / v[v!=0]
    h = np.zeros_like(v)
    med_v = v - med_v
    
    for i in range(3):
        loc = np.where((v==data[:, :, i]) & (med_v!=0))
        h[loc] = (data[:, :, (i+1)%3][loc] - data[:, :, (i+2)%3][loc]) / med_v[loc] + i * 2
    h[h<0] += 6
    
    # 调整饱和度s
    if type(scale).__name__ != 'function':
        s *= scale
    else:
        s = scale(s, **param)
    s[s>1] = 1
    
    f = h.copy()
    h = h.astype(int) % 6
    f -= h
    v = [v * (1 - s), v * (1 - (1 - f) * s), v, v * (1 - f * s)]
    
    # 初始化
    r = np.zeros_like(h)
    g = np.zeros_like(h)
    b = np.zeros_like(h)
    
    # 分别处理h为偶数和奇数时的rgb
    for i in range(3):
        loc = np.where(h==2 * i)
        r[loc] = v[(i+2)%3][loc]
        g[loc] = v[(i+1)%3][loc]
        b[loc] = v[i%3][loc]
    del v[1]
    for i in range(3):
        loc = np.where(h==2 * i + 1)
        r[loc] = v[(i+2)%3][loc]
        g[loc] = v[(i+1)%3][loc]
        b[loc] = v[i%3][loc]
    
    data = np.array([r,g,b], dtype=dtype)
    return data.transpose(1, 2, 0)


'''
本库用于科学计算和快速建模

maysics主要包括十二个模块：

1、algo 封装了几种模拟方法，用于简易模拟；
2、calc 封装了部分常见的算符算子和积分方法，辅助数学运算；
3、equa 封装了部分方程求解运算；
4、explain 用于评估和解释模型；
5、graph 用于图论分析；
6、image 用于数字图像处理;
7、models 封装了几种常用的模型以便快速构建数理模型；
8、prep 用于数据预处理；
9、stats 用于统计分析；
10、time 用于处理时间数据；
11、trans 储存了常用的坐标转换及其他数学变换；
12、utils 是额外工具箱。


This package is used for scientific calculating and fast modeling.

maysics includes twelve modules:

1. "algo" packages several simulation methods for simple simulation;
2. "calc" packages some common operators and integration method to assist in mathematical operations;
3. "equa" packages some equation solving operation;
4. "explain" is used for estimating and explaining model;
5. "graph" is used for graph theory analysis;
6. "image" is used for digital image process;
7. "models" packages several commonly used models for fast modeling;
8. "prep" is used for data preproccessing;
9. "stats" is uesd for statistical analysis;
10. "time" is used for processing time data;
11. "trans" stores common coordinate transformations and other mathematical transformations;
12. "utils" is extra Utils.
'''
import numpy as np
from scipy.special import factorial
import pickle, csv
from PIL import Image
import random
from . import algo, calc, equa, explain, graph, image,\
    models, prep, stats, time, trans, utils
from .models import linear_r
from .prep import preview, preview_file, shuffle, standard, minmax, normalizer
from .utils import circle, discrete, grid_net, hermit, mat_exp


# math
pi = 3.141592653589793
e = 2.718281828459045
golden = 1.618033988749895
gamma = 0.57721566490153286060651209
K = 0.76422365358922066299069873125
chaos_1 = 4.669201609102990
chaos_2 = 2.502907875095892
K0 = 2.6854520010


# physics
atm = 101325
c = 299792458.0
c_e = 1.602176634e-19
m_e = 9.1093837015e-31
m_earth = 5.965e24
m_n = 1.67492749804e-27
m_p = 1.67262192369e-27
m_sun = 1.9891e30
G = 6.6743e-11
g = 9.80665
h = 6.62607015e-34
hr = 1.0545718176461565e-34
k = 1.380649e-23
miu = 1.2566370614359173e-06
epsilon = 8.85e-12
NA = 6.02214179e23
R = 8.31446114959365
r_earth = 6371393
r_sun = 6.955e8
r_moon = 384403000
lambdac = 2.4263102175e-12
ly = 9460730472580800
AU = 149597870700
pc = 3.0856775814671915808e16
Platonic_year = 26000
SB = 5.670367e-8
v1 = 7900
v2 = 11200
v3 = 16700


def all_same(x):
    '''
    判断数组元素是否全相同
    
    参数
    ----
    x：数组
    
    返回
    ----
    布尔类型，True或者False
    
    
    Determine whether the array elements are all the same
    
    Parameter
    ---------
    x: array
    
    Return
    ------
    bool, True or False
    '''
    x = np.array(x)
    if len(x.shape) == 1:
        x = len(set(x))
        if x == 1:
            return True
        else:
            return False
    
    else:
        for i in x:
            if i.all() != x[0].all():
                return False
        return True


def choice(seq, pro=None, random_state=None):
    '''
    按指定概率抽取元素
    
    参数
    ----
    seq：一维列表，待抽取的元素
    pro：一维数组，可选，抽取相应元素的概率，默认概率全部相等
    random_state：整型，可选，随机种子
    
    返回
    ----
    seq中的元素
    
    
    Chooce elements according to the specified probability
    
    Parameters
    ----------
    seq: 1D array, elements to be chooced
    pro: 1D array, callable, the probability of choocing each element, the probability is equal by default
    random_state: int, callable, random seed
    
    Return
    ------
    the element in seq
    '''
    if not random_state is None:
        random.seed(random_state)
    
    if pro is None:
        return random.choice(seq)
    else:
        num_pro = len(pro)
        for i in range(1, num_pro):
            pro[i] += pro[i-1]
        random_num = random.random()
        for i in range(num_pro):
            if pro[i] >= random_num:
                return seq[i]


def covs1d(a, b, n):
    '''
    一维序列卷积和
    
    参数
    ----
    a：一维数组
    b：一维数组
    n：整型，平移步数
    
    返回
    ----
    数类型，a[n] * b[n]
    
    Convolution Sum of 1D List
    
    Parameters
    ----------
    a: 1D array
    b: 1D array
    n: int, translation steps
    
    Return
    ------
    num, a[n] * b[n]
    '''
    a = np.array(a)
    b = list(b)
    b.reverse()
    b = np.array(b)
    num_a = len(a)
    num_b = len(b)
    if n <= 0 or n >= num_a + num_b:
        result = 0
    
    else:
        a = np.hstack((np.zeros(num_b), a, np.zeros(num_b)))
        b = np.hstack((b, np.zeros(num_a + num_b)))
        b[n : n+num_b] = b[: num_b]
        b[: n] = 0
        result = sum(a * b)
    
    return result


def covs2d(a, b, n, m):
    '''
    二维序列卷积和
    
    参数
    ----
    a：二维数组
    b：二维数组
    n：整型，沿axis=0方向的平移步数
    m：整型，沿axis=1方向的平移步数
    
    返回
    ----
    数类型，a[n, m] * b[n, m]
    
    Convolution Sum of 2D List
    
    Parameters
    ----------
    a: 2D array
    b: 2D array
    n: int, translation steps along axis=0
    m: int, translation steps along axis=1
    
    Return
    ------
    num, a[n, m] * b[n, m]
    '''
    a = np.array(a)
    b = np.array(b)
    b = np.fliplr(b)
    b = np.flipud(b)
    num_a_x = a.shape[1]
    num_a_y = a.shape[0]
    num_b_x = b.shape[1]
    num_b_y = b.shape[0]
    
    if n <= 0 and m <= 0 or n >= num_a_x + num_b_x and m >= num_a_y + num_b_y:
        result = 0
    
    else:
        a = np.hstack((np.zeros((num_a_y, num_b_x)), a, np.zeros((num_a_y, num_b_x))))
        a = np.vstack((np.zeros((num_b_y, num_a_x + 2 * num_b_x)), a, np.zeros((num_b_y, num_a_x + 2 * num_b_x))))
        b = np.hstack((b, np.zeros((num_b_y, num_a_x + num_b_x))))
        b = np.vstack((b, np.zeros((num_a_y + num_b_y, num_a_x + 2 * num_b_x))))
        
        # 移动b矩阵
        b[n : n+num_b_y] = b[: num_b_y]
        b[: n] = 0
        b[:, m : m+num_b_x] = b[:, : num_b_x]
        b[:, : m] = 0
        
        result = (a * b).sum()
    
    return result


def load(filename, header=True, pic=False, dtype='uint8'):
    '''
    载入pkl、npy、csv文件或图片
    
    参数
    ----
    filename：字符串类型，文件名
    header：布尔类型，可选，True表示csv文件第一行为列名，仅在读取csv文件时有效，默认为True
    pic：布尔类型，可选，True表示读取图片，默认为False
    dtype：可选，输出图像数据类型，仅在pic=True时有效，默认为'uint8'
    
    
    Load pkl, npy, csv file or picture
    
    Parameter
    ---------
    filename: str, file name
    header: bool, callable, True means the first row of the csv file if the names of columns, effective only when reading csv files, default=True
    pic: bool, callable, True means to load picture, default=False
    dtype: callable, data format of output image, effective only when pic=True, default='uint8'
    '''
    if pic is False:
        if filename[-4:] == '.pkl':
            with open(filename, 'rb') as file:
                data = pickle.load(file)
            
            return data
        
        elif filename[-4:] == '.npy':
            return np.load(filename, allow_pickle=True)
        
        elif filename[-4:] == '.csv':
            with open(filename, 'r', encoding='utf-8') as f:
                reader = list(csv.reader(f))
                if header:
                    reader = reader[1:]
                return np.array(reader)
        
        else:
            raise Exception("Suffix of filename must be '.pkl', '.npy' or '.csv'.")
    
    else:
        x = Image.open(filename)
        return np.asarray(x, dtype=dtype)


def save(filename, data, header=None, pic=False):
    '''
    保存为pkl、npy、csv文件或图片
    
    参数
    ----
    filename：字符串类型，文件名
    data：需要保存的数据
    header：一维列表类型，可选，数据的列名称，仅在写入csv文件时有效
    pic：布尔类型，可选，True表示保存为图片，默认为False
    
    
    Save as pkl, npy, csv file or picture
    
    Parameters
    ----------
    filename: str, file name
    data: data
    header: 1D list, callable, the names of columns, effective only when writing csv files
    pic: bool, callable, True means to save as picture, default=False
    '''
    if pic is False:
        if filename[-4:] == '.pkl':
            with open(filename, 'wb') as file:
                pickle.dump(data, file)
        
        elif filename[-4:] == '.npy':
            np.save(filename, data)
        
        elif filename[-4:] == '.csv':
            data = np.array(data, dtype=np.object)
            if not header:
                header = []
                if len(data.shape) == 1:
                    for i in range(data.shape[0]):
                        header.append(i)
                        with open(filename, 'w', newline='') as f:
                            writer = csv.writer(f)
                            writer.writerow(header)
                            writer.writerow(data)
                
                else:
                    for i in range(data.shape[1]):
                        header.append(i)
                        with open(filename, 'w', newline='') as f:
                            writer = csv.writer(f)
                            writer.writerow(header)
                            writer.writerows(data)
        
        else:
            raise Exception("Suffix of filename must be '.pkl', '.npy' or '.csv'.")
    
    else:
        data = np.array(data, dtype='uint8')
        image = Image.fromarray(data)
        image.save(filename)


def lag(x, n, m):
    '''
    （连带）拉盖尔多项式
    
    参数
    ----
    x：数，输入值
    n：整型，次数
    m：整型，阶数
    
    
    (Associated) Laguerre Polynomial
    
    Parameters
    ----------
    x: num, input value
    n: int, degree
    m: int, order
    '''
    result = 0
    for i in range(n+1):
        result += (-1)**i * factorial(n + m) / (factorial(n - i) * factorial(i + m) * factorial(i)) * x**i
    return result


def lpn(l):
    '''
    l阶勒让德多项式的模
    
    参数
    ----
    l：整型，阶数
    
    返回
    ----
    数类型
    
    
    Module of Legendre Polynomials at degree l
    
    Paramater
    ---------
    l: int, degree
    
    Return
    ------
    num
    '''
    return (2 / (2 * l + 1))**0.5


def lp(x, l):
    '''
    l阶勒让德多项式的值
    
    参数
    ----
    x：数，输入值
    l：整型，阶数
    
    返回
    ----
    数
    
    
    Value of Legendre Polynomials at degree l
    
    Parameters
    ----------
    x: num, input value
    l: int, degree
    
    Return
    ------
    num
    '''
    result = 0
    for k in range(int(l/2) + 1):
        result += (-1)**k * factorial(2 * l - 2 * k)\
        / 2**l / factorial(k) / factorial(l - k) / factorial(l - 2 * k)\
        * x**(l - 2 * k)
    return result


def alpn(l, m):
    '''
    m阶l次连带勒让德多项式的模
    
    参数
    ----
    l：整型，次数
    m：整型，阶数
    
    Value of Associated Legendre Polynomials of order m and degree l
    
    Parameters
    ----------
    l: int, degree
    m: int, order
    '''
    up_part = 2 * factorial(l + m)
    down_part = (2 * l + 1) * factorial(l - m)
    return (up_part / down_part)**0.5


def _dm_lp(x, l, m, acc):
    '''
    勒让德多项式的m阶导数
    '''
    if m == 0:
        x = lp(x, l)
    elif m == 1:
        x = (lp(x + acc * 0.5, l) - lp(x - acc * 0.5, l)) / acc
    else:
        x = (_dm_lp(x + acc * 0.5, l, m-1, acc) - _dm_lp(x - acc * 0.5, l, m-1, acc)) / acc
    return x


def alp(x, l, m, acc=0.1):
    '''
    m阶l次连带勒让德多项式的值
    Plm(x) = (1-x^2)^(m/2) * d^m/dx^m * Pl(x)
    
    参数
    ----
    x：数，输入值
    l：整型，次数
    m：整型，阶数
    acc：浮点数类型，求导精度
    
    
    Value of Associated Legendre Polynomials of order m and degree l
    Plm(x) = (1-x^2)^(m/2) * d^m/dx^m * Pl(x)
    
    Parameters
    ----------
    x: num, input value
    l: int, degree
    m: int, order
    acc: float, callable, accuracy of derivation, default=0.1
    '''
    return (1 - x**2)**(m / 2) * _dm_lp(x, l, m, acc)


def hp(x, n, li=False):
    '''
    n阶厄米多项式

    参数
    ----
    x：数，输入值
    n：整型，阶数
    li：布尔类型，可选，True表示将0到n阶勒让德多项式组成一维ndarray返回，默认为False

    返回
    ----
    数或一维ndarray


    Value of Hermite Polynomials at degree n

    Parameters
    ----------
    x: num, input value
    n: int, degree
    li: bool, callable, True means that 0 to n order Legendre polynomials are formed into 1D ndarray and returned, default=False
    
    Reuturn
    -------
    num or 1D ndarray
    '''
    if n <= 0:
        n = 0
    
    if not li:
        if n == 0:
            return 1

        elif n == 1:
            return 2 * x

        else:
            return 2 * x * hp(x, n-1) - 2 * (n-1) * hp(x, n-2)
    
    else:
        hp_list = []
        for i in range(n+1):
            if i == 0:
                hp_list.append(1)
            
            elif i == 1:
                hp_list.append(2 * x)
            
            else:
                hp_list.append(2 * x * hp_list[i-1] - 2 * (i-1) * hp_list[i-2])
        return np.array(hp_list)


def v_mean(m, T):
    '''
    麦克斯韦速率分布律下的平均速率
    
    参数
    ----
    m：气体分子质量, 单位：kg
    T：气体温度, 单位：K
    
    返回
    ----
    数或数组类型，速率，单位：m/s
    
    
    Average Velocity in Maxwell's Velocity Distribution Law
    
    Parameters
    ----------
    m: mass of gas molecule, unit: kg
    T: temperature of gas, unit: K
    
    Return
    ------
    num or array, velocity, unit: m/s
    '''
    return 8**0.5 * (k * T / (np.pi * m))**0.5


def v_p(m, T):
    '''
    麦克斯韦速率分布律下的最概然速率
    
    参数
    ----
    m：气体分子质量, 单位：kg
    T：气体温度, 单位：K
    
    返回
    ----
    数或数组类型，速率，单位：m/s
    
    
    Most Probable Velocity in Maxwell's Velocity Distribution Law
    
    Parameters
    ----------
    m: mass of gas molecule, unit: kg
    T: temperature of gas, unit: K
    
    Return
    ------
    num or array, velocity, unit: m/s
    '''
    return 2**0.5 * (k * T / (np.pi * m))**0.5


def v_rms(m, T):
    '''
    麦克斯韦速率分布律下的均方根速率
    
    参数
    ----
    m：气体分子质量, 单位：kg
    T：气体温度, 单位：K
    
    返回
    ----
    数或数组类型，速率，单位：m/s
    
    
    Root-Mean-Square Velocity Maxwell's Velocity Distribution Law
    
    Parameters
    ----------
    m: mass of gas molecule, unit: kg
    T: temperature of gas, unit: K
    
    Return
    ------
    num or array, velocity, unit: m/s
    '''
    return 3**0.5 * (k * T / (np.pi * m))**0.5


'''
本模块是额外工具箱

This module is extra Utils
'''
import numpy as np
from urllib import request
from lxml import etree
from urllib import parse
import string
from matplotlib import pyplot as plt
from maysics import trans


def grid_net(*args):
    '''
    生成网格点
    将输入的列表遍历组合
    如：
    grid_net([a, b], [c, d]) 或 grid_net(*[[a, b], [c, d]])
    返回：array([[a, c], [b, c], [a, d], [b, d]])
    
    
    Generate grid
    traverse and combine the input list
    e.g.
    grid_net([a, b], [c, d]) or grid_net(*[[a, b], [c, d]])
    return: array([[a, c], [b, c], [a, d], [b, d]])
    '''
    net = np.meshgrid(*args)
    for i in range(len(net)):
        net[i] = net[i].flatten()
    net = np.vstack(tuple(net)).T
    return net


def hermit(data):
    '''
    求厄密共轭矩阵
    
    参数
    ----
    data：二维数组
    
    返回
    ----
    二维ndarray
    
    
    Hermitian Conjugate Matrix
    
    Parameter
    ---------
    data: 2-D array
    
    Return
    ------
    2-D ndarray
    '''
    data = np.array(data)
    return data.T.conj()


def mat_exp(x, tol=0.01):
    '''
    矩阵的exp运算
    求e^data
    
    参数
    ----
    x：二维数组，矩阵
    tol：浮点数类型，可选，误差，当data^n/n!的每个矩阵元均小于tol时输出结果
    
    返回
    ----
    二维ndarray
    
    
    Exp Operation of Matrix
    
    Parameters
    ----------
    x: 2-D array, matrix
    tol: float, callable, error, output when every element of data^n/n! less than tol
    
    Return
    ------
    2-D ndarray
    '''
    x = np.matrix(x)
    I_x = np.matrix(np.eye(x.shape[0]))
    result_up = I_x.copy()
    error = I_x.copy()
    result = I_x.copy()
    n = 0
    result_down = 1
    while (error > tol).any():
        n += 1
        result_up *= x
        result_down *= n
        error = result_up / result_down
        result += error
    return np.array(result)


def e_distance(p1, p2):
    '''
    求某两个点之间的欧式距离
    
    参数
    ----
    p1：一维数组，第一个点的位置
    p2：一维数组，第二个点的位置
    
    返回
    ---
    浮点数类型，距离
    
    
    Calculate the Euclidean distance between two points
    
    Parameters
    ----------
    p1: 1-D list, the location of the first point
    p2: 1-D list, the location of the second point
    
    Return
    ------
    float, the distance
    '''
    p1 = np.array(p1)
    p2 = np.array(p2)
    
    return sum((p1 - p2)**2)**0.5


def e_distances(data, des='o'):
    '''
    求data到目标点的欧式距离
    
    参数
    ----
    data：一维或二维列表，数据
    des：字符串或一维数组，可选'o'或'O'(原点)、'mean'(均值点)及自定义数组，目标点坐标，默认为'o'
    
    返回
    ----
    ndarray类型，距离数组
    
    
    Calculate the Euclidean distances between data and destination
    
    Parameter
    ---------
    data: 1-D or 2-D list, data
    des: str or 1-D list, 'o' or 'O' (origin), 'mean' (mean point) and custom array are optional, the coordinate of destination, default='o'
    
    Return
    ------
    ndarray, the distances
    '''
    data = np.array(data, dtype=np.float)
    if len(data.shape) < 3:
        n = data.shape[-1]
    else:
        raise Exception("Parameter 'data' must be 1-D or 2-D.")

    if des == 'o' or des == 'O':
        des = np.zeros(n)

    elif des == 'mean':
        des = data.mean(axis=0)
        
    else:
        des = np.array(des)
    
    data = (data - des)**2
    result = data.sum(axis=len(data.shape)-1)
    result = result**0.5
    
    return result


def m_distance(data, p1, p2):
    '''
    求某两个点之间的马氏距离
    
    参数
    ----
    data：二维列表，数据
    p1：一维或二维数组，第一个点的位置
    p2：一维或二维数组，第二个点的位置
    
    返回
    ---
    浮点数类型，距离
    
    
    Calculate the Mahalanobis distance between two points
    
    Parameters
    ----------
    data: 2-D list, data
    p1: 1-D or 2-D list, the location of the first point
    p2: 1-D or 2-D list, the location of the second point
    
    Return
    ------
    float, the distance
    '''
    data = np.mat(data, dtype=np.float)
    dataT = data.T
    
    if len(data.shape) != 2:
        raise Exception("Parameter 'data' must be 2-D.")
    
    SI = np.mat(np.cov(dataT)).I

    p1 = np.mat(p1, dtype=np.float)
    p2 = np.mat(p2, dtype=np.float)
    result = (p1 - p2) * SI * (p1 - p2).T
    
    return result[0, 0]**0.5


def m_distances(data, des='o'):
    '''
    求data到目标点的马氏距离
    
    参数
    ----
    data：二维列表，数据
    des：字符串或一维或二维数组，可选'o'或'O'(原点)、'mean'(均值点)及自定义数组，目标点坐标，默认为'o'
    
    返回
    ----
    ndarray类型，距离数组
    
    
    Calculate the Mahalanobis distance between data and destination
    
    Parameter
    ---------
    data: 2-D list, data
    des: str or 1-D or 2-D list, 'o' or 'O' (origin), 'mean' (mean point) and custom array are optional, the coordinate of destination, default='o'
    
    Return
    ------
    ndarray, the distances
    '''
    data = np.mat(data, dtype=np.float)
    dataT = data.T
    
    if len(data.shape) != 2:
        raise Exception("Parameter 'data' must be 2-D.")
    
    SI = np.mat(np.cov(dataT)).I

    if des == 'o' or des == 'O':
        des = np.zeros(data.shape[-1])
    
    elif des == 'mean':
        des = data.mean(axis=0)
    
    else:
        des = np.mat(des)
    
    return np.diag((data - des) * SI *(dataT - des.T))**0.5


def discrete(x, y, color=None, label=None):
    '''
    绘制离散函数图像
    
    参数
    ----
    x：一维数组类型，自变量
    y：一维数组类型，因变量
    color：字符串类型，可选，颜色
    label：字符串类型，可选，标签
    
    
    Draw the graph of discrete function
    
    Parameters
    ----------
    x: 1-D array, independent variable
    y: 1-D array, dependent variable
    color: str, callable, color
    label: str, callable, label
    '''
    plt.scatter(x, y, color=color, label=label)
    zeros = np.zeros_like(y)
    if not color:
        plt.vlines(x, zeros, y)
    else:
        plt.vlines(x, zeros, y, color=color)


def circle(center=(0, 0), radius=1, angle_range=(0, 2*np.pi), acc=0.01, c=None, label=None):
    '''
    绘制一个圆
    
    参数
    ----
    center：元组类型，可选，圆心坐标，默认为(0, 0)
    radius：数类型，可选，半径，默认为1
    angle_range：元组类型，可选，绘制的角度范围，默认为(0, 2π)
    acc：浮点数类型，可选，绘制的精度，默认为0.01
    c：可选，颜色
    label：字符串类型，可选，标签，默认为None
    
    
    Draw a circle
    
    Parameters
    ----------
    center: tuple, callable, center coordinate, default=(0, 0)
    radius: num, callable, radius, default=1
    angle_range: tuple, callable, the range of angle to draw, default=(0, 2π)
    acc: float, callable, the accuracy of drawing, default=0.01
    c: callable, color
    label: str, callable, label, default=None
    '''
    theta = np.arange(*angle_range, acc)
    radius = radius * np.ones_like(theta)
    x = np.vstack((radius, theta)).T
    x = trans.ipolar(x)
    plt.plot(x[:, 0] + center[0], x[:, 1] + center[1], c=c)


class A_P():
    '''
    将信号的频域表示分解为“幅度-频率”和“相位-频率”

    参数
    ----
    X：函数或一维数组形式，信号的频域表示


    Decompose frequency domain representation of signal into "amplitude-frequency" and "phase-frequency"

    Parameter
    ---------
    X: function or 1-D array, frequency domain representation of signal
    '''
    def __init__(self, X):
        self.X = X
    

    def fit(self, f, param={}):
        '''
        计算频率为f时的幅度和相位

        参数
        ----
        f：数或一维数组形式，频率

        返回
        ----
        元组形式，(幅度, 相位)


        Calculate the amplitude and phase at frequency f

        Parameter
        ---------
        f: num or 1-D array, frequency

        Return
        ------
        tuple, (amplitude, phase)
        '''
        self.f = np.array(f, np.float)
        f = self.f.astype(complex)
        if type(self.X).__name__ == 'function':
            result = self.X(f, **param)
        else:
            X = np.array(self.X)
            result = self.X[f]
        result = np.array(result)
        self.amplitude = abs(result)
        
        index1 = np.where(result.imag == 0)[0]
        index2 = np.where(result.imag != 0)[0]
        result_new = result[index1]
        result_new[result_new.real == 0] = 0
        result_new[result_new.real > 0] = np.pi / 2
        result_new[result_new.real < 0] = -np.pi / 2
        result[index1] = result_new
        result_new = result[index2]
        result[index2] = np.arctan(result_new.imag / result_new.real)
        self.phase = result.real
    

    def __image_process(self, image_type):
        fig = plt.figure()
        if image_type == 'C' or image_type == 'c':
            ax = fig.add_subplot(2, 1, 1)
            ax.plot(self.f, self.amplitude)
            ax.set_title('amplitude')
            ax = fig.add_subplot(2, 1, 2)
            ax.plot(self.f, self.phase)
            ax.set_title('phase')
        elif image_type == 'D' or image_type == 'd':
            zeros_list = np.zeros(self.f.shape)
            ax = fig.add_subplot(2, 1, 1)
            ax.scatter(self.f, self.amplitude, marker='o', s=30, zorder=3)
            ax.vlines(self.f, zeros_list, self.amplitude)
            ax.set_title('amplitude')
            ax = fig.add_subplot(2, 1, 2)
            ax.scatter(self.f, self.phase, marker='o', s=30, zorder=3)
            ax.vlines(self.f, zeros_list, self.phase)
            ax.set_title('phase')
        plt.tight_layout()


    def show(self, image_type='c'):
        '''
        显示“幅度-频率”图和“相位-频率”图

        参数
        ----
        image_type：字符串形式，可选'c'和'd'，'c'表示绘制连续图像，'d'表示绘制离散图像，默认为'd'


        Display "amplitude-frequency" and "phase-frequency" graphs

        Parameter
        ---------
        image_type: str, 'c' and 'd' are callable, 'c' means drawing continuous image and 'd'means drawing discrete image, default='c'
        '''
        self.__image_process(image_type)
        plt.show()
    

    def savefig(self, filename, image_type='c'):
        '''
        储存“幅度-频率”图和“相位-频率”图

        参数
        ----
        filename：字符串形式，文件名
        image_type：字符串形式，可选'C'和'D'，'C'表示绘制连续图像，'D'表示绘制离散图像，默认为'C'


        Save "amplitude-frequency" and "phase-frequency" graphs

        Parameters
        ----------
        filename: str, file name
        image_type: str, 'c' and 'd' are callable, 'c' means drawing continuous image and 'd'means drawing discrete image, default='c'
        '''
        self.__image_process(image_type)
        plt.savefig(filename)


'''
本模块储存着部分常用的模型

This module stores some commonly used models
'''
import numpy as np
from scipy.integrate import solve_ivp
from scipy.sparse import csr_matrix


def linear_r(x, y):
    '''
    线性回归模型
    
    参数
    ----
    x：一维或二维ndarray，自变量
    y：一维或二维ndarray，因变量
    
    返回
    ----
    一维ndarray：系数行矩阵
    损失函数：平均平方误差mse
    
    
    Linear Regression Model
    
    Parameters
    ----------
    x: 1D or 2D ndarray, independent variable
    y: 1D or 2D ndarray, dependent variable
    
    Return
    ------
    1D ndarray: coefficient row matrix
    loss: mean squared error: mse
    '''
    x = np.array(x)
    if len(x.shape) == 1:
        x = np.array([x]).T
    y = np.array(y)
    x = np.hstack((x, np.ones((x.shape[0], 1))))
    coef = np.linalg.lstsq(x, y, rcond=None)[0].reshape(-1)
    y_pre = np.dot(x, np.array([coef]).T).reshape(y.shape)
    mse = ((y_pre - y)**2).mean()
    return coef, mse


def logistic(t, N0, r, K):
    '''
    Logistic人口增长模型
    该模型可得到解析解
    解的形式为：Nt = K * N0 / (N0 + (K - N0) * e**(-r * t))
    其中，Nt为t时刻的人口数
    
    参数
    ----
    t：时间
    N0：数或数组，现有人口数
    r：数类型，人口自然增长率
    K：数类型，环境资源允许的稳定人口数
    
    返回
    ----
    数，Nt
    
    
    Logistic population growth models
    solution: Nt = K * N0 / (N0 + (K - N0) * e**(-r * t))
    Nt is the population at time 't'
    
    Parameters
    ----------
    t: time
    N0: num or array, initial population
    r: num, natural population growth rate
    K: num, stable population allowed by environmental resources
    
    Return
    ------
    num, Nt
    '''
    Nt_pre_down_ = N0 + (K - N0) * np.e**(-r * t)
    Nt = K * N0 / Nt_pre_down_
    return Nt


class ED():
    '''
    传染病模型
    本类可构建SI模型、SIR模型、SIRS模型、SEIR模型
    S是易感染者
    E是暴露者，染病但还不具有传染能力
    I是感染者，带有传染能力
    R是康复者(带有免疫力)或死者
    
    参数
    ----
    I0：数类型，初始感染者
    K：数类型，样本总数
    beta：数类型，感染率
    R：数类型，可选，初始康复者(带有免疫力)或死者，默认为0
    E：数类型，可选，暴露者，默认为0
    
    
    mathematical models of epidemic diseases
    this class is used to buil SI, SIR, SIRS, SEIR
    S means susceptible people
    E means exposed people, which are infected without ability to infect
    I means infected people with ability to infect
    R means recovered people with immunity or dead people
    
    Parameters
    ----------
    I0: initial infected people
    K: sample size
    beta: infection rate
    R: num, callable, recovered people with immunity or dead people, default=0
    E: num, callable, exposed people, default=0
    '''
    def __init__(self, I0, K, beta, R=0, E=0):
        self.I0 = I0
        self.K = K
        self.beta = beta
        self.R = R
        self.E = E
        self.S = K - I0 - R - E
    
    
    def SI(self, t):
        '''
        SI模型：
        该模型不需要再额外输入参数，且可得到解析解
        解的形式为：I = K*(1-(K-I0)/(K+I0+I0*e**(beta*K*t)))
        
        参数
        ----
        t：数或一维数组，需要预测的时间

        返回
        ----
        元组，(S(t), I(t))
        
        
        SI:
        no more parameters
        solution: I = K*(1-(K-I0)/(K+I0+I0*np.e**(beta*K*t)))
        
        Parameter
        ---------
        t: num or 1D array, time for prediction
        
        Return
        ------
        tuple, (S(t), I(t))
        '''
        I_pre_down_ = self.K + self.I0 * (np.e**(self.beta * self.K * t) - 1)
        I_pre_up_ = self.K - self.I0
        I_func_ = self.K * (1 - I_pre_up_ / I_pre_down_)
        S_func_ = self.K - I_func_
        return S_func_, I_func_
    
    
    def SIR(self, gamma, t_span, method='RK45', t_eval=None):
        '''
        SIR模型：
        该模型不能得到解析解，给出的是基于solve_ivp得到的数值解数组
        
        参数
        ----
        gamma：数类型，康复率
        t_span：元组形式，求解的上下限
        method：字符串形式，可选，求解方法，可选择'RK45'、'RK23'、'DOP835'、'Radau'、'BDF'、'LSODA'
        t_eval：数组形式，可选，每当t等于该数组中的值时，会生成一个数值解

        返回
        ----
        solve_ivp数值解，顺序是S, I, R
        
        
        SIR:
        
        Parameters
        ----------
        gamma: num, recovery rate
        t_span: tuple, limits of solution
        method: str, callable, solving method, 'RK45', 'RK23', 'DOP835', 'Radau', 'BDF', 'LSODA' are optional
        t_eval: list, callable, when 't' equals the value in the list, it will generate a numerical solution
        
        Return
        ------
        numerical solution of solve_ivp, the order is S, I, R
        '''
        y0 = [self.S, self.I0, self.R]
        def epis_equas(t, x):
            y1 = -self.beta * x[1] * x[0]
            y2 = self.beta * x[1] * x[0] - gamma * x[1]
            y3 = gamma * x[1]
            return y1, y2, y3
        result = solve_ivp(epis_equas, t_span=t_span, y0=y0, method=method, t_eval=t_eval)
        return result.y
    
    
    def SIRS(self, gamma, alpha, t_span, method='RK45', t_eval=None):
        '''
        SIRS模型：
        该模型不能得到解析解，给出的是基于solve_ivp得到的数值解数组
        
        参数
        ----
        gamma：数类型，康复率
        alpha：衡量康复者获得免疫的时间
        t_span：元组形式，求解的上下限
        method：字符串形式，可选，求解方法，可选择'RK45'、'RK23'、'DOP835'、'Radau'、'BDF'、'LSODA'
        t_eval：数组形式，可选，每当t等于该数组中的值时，会生成一个数值解

        返回
        ----
        solve_ivp数值解，顺序是S, I, R
        
        
        SIRS:
        
        Parameters
        ----------
        gamma: num, recovery rate
        alpha: measuring the time that the recovered people with immunity
        t_span: tuple, limits of solution
        method: str, callable, solving method, 'RK45', 'RK23', 'DOP835', 'Radau', 'BDF', 'LSODA' are optional
        t_eval: list, callable, when 't' equals the value in the list, it will generate a numerical solution
        
        Return
        ------
        numerical solution of solve_ivp, the order is S, I, R
        '''
        y0= [self.S, self.I0, self.R]
        def epis_equas(t, x):
            y1 = -self.beta * x[1] * x[0] + alpha * x[2]
            y2 = self.beta * x[1] * x[0] - gamma * x[1]
            y3 = gamma * x[1] - alpha * x[2]
            return y1, y2, y3
        result = solve_ivp(epis_equas, t_span=t_span, y0=y0, method=method, t_eval=t_eval)
        return result.y
    
    
    def SEIR(self, gamma1, gamma2, alpha, t_span, method='RK45', t_eval=None):
        '''
        SEIR模型：
        该模型不能得到解析解，给出的是基于solve_ivp得到的数值解数组
        
        参数
        ----
        gamma1：数类型，潜伏期康复率
        gamma2：数类型，患者康复率
        alpha：数类型，衡量康复者获得免疫的时间
        t_span：元组形式，求解的上下限
        method：字符串形式，可选，求解方法，可选择'RK45'、'RK23'、'DOP835'、'Radau'、'BDF'、'LSODA'
        t_eval：数组形式，可选，每当t等于该数组中的值时，会生成一个数值解

        返回
        ----
        solve_ivp数值解，顺序是S, E, I, R
        
        
        SEIR:
        
        Parameters
        ----------
        gamma1: num, recovery rate of incubation
        gamma2: num, recovery rate of patients
        alpha: num, measuring the time that the recovered people with immunity
        t_span: tuple, limits of solution
        method: str, callable, solving method, 'RK45', 'RK23', 'DOP835', 'Radau', 'BDF', 'LSODA' are optional
        t_eval: list, callable, when 't' equals the value in the list, it will generate a numerical solution
        
        Return
        ------
        numerical solution of solve_ivp, the order is S, E, I, R
        '''
        y0 = [self.S, self.E, self.I0, self.R]
        def epis_equas(t, x):
            y1 = -self.beta * x[2] * x[0]
            y2 = self.beta * x[2] * x[0] - (alpha + gamma1) * x[1]
            y3 = alpha * x[1] - gamma2 * x[2]
            y4 = gamma1 * x[1] + gamma2 * x[2]
            return y1, y2, y3, y4
        result = solve_ivp(epis_equas, t_span=t_span, y0=y0, method=method, t_eval=t_eval)
        return result.y


class Leslie():
    '''
    Leslie模型
    解的形式为：Nt = M**t * N0
    其中，Nt是t时刻的个体数列表，M是Leslie矩阵
    
    参数
    ----
    N0：一维数组，各年龄层初始个体数目
    r：一维数组，各年龄层的生殖率
    s：一维数组，各年龄层到下一个年龄层的存活率
    age_range：整型，年龄段的跨度，默认为1
    
    属性
    ----
    Leslie_matrix：莱斯利矩阵
    
    
    Leslie model
    solution: Nt = M**t * N0
    Nt is the list of individuals at time 't' , M is Leslie matrix
    
    Parameters
    ----------
    N0: 1D array, initial number of individuals in each age group
    r: 1D array, reproductive rate of each age group
    s: 1D array, survival rate to next age group of each group
    age_range: int, the span of age groups, default=1
    
    Attribute
    ---------
    Leslie_matrix: Leslie matrix 
    '''
    def __init__(self, N0, r, s, age_range=1):
        self.N0 = np.array(N0)
        self.r = np.array(r)
        self.s = np.array(s)
        self.age_range = age_range
        
        Leslie_matrix = np.zeros((len(s), len(N0)))
        Leslie_matrix = np.vstack((r, Leslie_matrix))
        for i in range(1, len(N0)):
            Leslie_matrix[i][i-1] = s[i-1]
        self.Leslie_matrix = csr_matrix(Leslie_matrix)
    
    
    def predict(self, t):
        '''
        参数
        ----
        t：数或一维数组，时间
        
        返回
        ----
        数，Nt
        
        
        Parameter
        ---------
        t: num or 1D array, time
        
        Return
        ------
        num, Nt
        '''
        t_times = t // self.age_range
        dense_Leslie_matrix = self.Leslie_matrix.todense()
        Nt = self.N0 * dense_Leslie_matrix**t_times
        return np.array(Nt)[0]